// $Header: /u/drspeech/repos/quicknet2/QN_fltvec.h,v 1.4 2004/05/20 23:38:20 davidj Exp $

// Vector floating point utility routines for QuickNet

#ifndef QN_fltvec_h_INCLUDED
#define QN_fltvec_h_INCLUDED

#include <stddef.h>
#include <math.h>
#include <QN_config.h>
#ifdef QN_HAVE_LIBCBLAS
extern "C" {
#include "cblas.h"
}
#endif
#include "QN_types.h"
#include "QN_intvec.h"

//// This global variable decides what type of optimization to use.

extern QNUInt32 qn_math;

//// These are in QN_fltvec.cc ////
// Convert a big-endian double vector into host format
void qn_btoh_vd_vd(size_t len, const double *from, double *to);
void qn_htob_vd_vd(size_t len, const double *from, double *to);
void qn_nv_sigmoid_vf_vf(size_t, const float* in_vec, float* out_vec);
void qn_pp_sigmoid_vf_vf(size_t, const float* in_vec, float* out_vec);
void qn_nv_softmax_vf_vf(size_t n, const float* in_vec, float* out_vec);
void qn_pp_softmax_vf_vf(size_t n, const float* in_vec, float* out_vec);
void qn_htopre8_vf_vi8(size_t len, const float* from, QNInt8* to);
void qn_pre8toh_vi8_vf(size_t len, const QNInt8* bits, float* res);
void qn_htolna8_vf_vi8(size_t len, const float* from, QNInt8* to);
void qn_lna8toh_vi8_vf(size_t len, const QNInt8* bits, float* res);
size_t qn_imax_vf_u(size_t n, const float* v);
float* qn_pre8table_vf(void);
float* qn_lna8table_vf(void);
// Fill each row of a matrix with the same vector.
void qn_copy_vf_mf(size_t mat_height, size_t vec_len, const float* vec,
		   float* mat);
// Sum the columns of a matrix into a vector.
void qn_sumcol_mf_vf(size_t rows, size_t cols, const float* in, float* res);

// Copy a dense matrix into a strided matrix.
void qn_copy_mf_smf(size_t height, size_t in_width, size_t stride,
		    const float* from, float* to);
// Copy a strided matrix into a dense matrix.
void qn_copy_smf_mf(size_t height, size_t width, size_t stride,
		    const float* from, float* to);
// Transopos a matrix.  Note that the height and width refer to the first 
// matrix - they're the other way around in the second.
void qn_trans_mf_mf(size_t in_height, size_t in_width, const float *in,
		    float *res);

// Stuff we need below
inline void
qn_copy_f_vf(size_t len, float from, float* to)
{
    size_t i;

    for (i=0; i<len; i++)
	*to++ = from;
}
    
// Convolve two vectors of floats
void qn_nv_convol_vfvf_vf(size_t h_len, size_t x_len, const float *h,
			  const float *x, float *y);
void qn_pp_convol_vfvf_vf(size_t h_len, size_t x_len, const float *h,
			  const float *x, float *y);
inline void
qn_convol_vfvf_vf(size_t h_len, size_t x_len, const float *h,
		       const float *x, float *y)
{
    if (qn_math & QN_MATH_PP)
	qn_pp_convol_vfvf_vf(h_len, x_len, h, x, y);
    else
	qn_nv_convol_vfvf_vf(h_len, x_len, h, x, y);
}

// Matrix-vector multiply (forward pass)
void qn_nv_mul_mfvf_vf(const size_t rows, const size_t cols,
			const float* mat, const float* vec, float* res);
void qn_pp_mul_mfvf_vf(const size_t rows, const size_t cols,
		       const float* mat, const float* vec, float* res);
inline void qn_mul_mfvf_vf(const size_t rows, const size_t cols,
			   const float* mat, const float* vec, float* res)
{
    if (qn_math & QN_MATH_PP)
	qn_pp_mul_mfvf_vf(rows, cols, mat, vec, res);
    else
	qn_nv_mul_mfvf_vf(rows, cols, mat, vec, res);
}

// Scaled vector outer product (weight update)
void qn_nv_mulacc_fvfvf_mf(const size_t rows, const size_t cols,
			const float scale, const float* in_vec,
			const float* in_vec_t, float *mat);
void qn_pp_mulacc_fvfvf_mf(const size_t rows, const size_t cols,
			const float scale, const float* in_vec,
			const float* in_vec_t, float *mat);
inline void
qn_mulacc_fvfvf_mf(const size_t rows, const size_t cols,
		   const float scale, const float* in_vec,
		   const float* in_vec_t, float *mat)
{
    if (qn_math & QN_MATH_PP)
	qn_pp_mulacc_fvfvf_mf(rows, cols, scale, in_vec, in_vec_t, mat);
    else
	qn_nv_mulacc_fvfvf_mf(rows, cols, scale, in_vec, in_vec_t, mat);
}

// Matrix-vector multiply - back propagation
void qn_nv_mul_vfmf_vf(const size_t rows, const size_t cols,
		    const float* vec, const float* mat, float* res);
void qn_pp_mul_vfmf_vf(const size_t rows, const size_t cols,
		    const float* vec, const float* mat, float* res);
inline void
qn_mul_vfmf_vf(const size_t rows, const size_t cols,
	       const float* vec, const float* mat, float* res)
{
    if (qn_math & QN_MATH_PP)
	qn_pp_mul_vfmf_vf(rows, cols, vec, mat, res);
    else
	qn_nv_mul_vfmf_vf(rows, cols, vec, mat, res);
}

// SAXYP - used by mulacc_fvfvf_mf
void qn_nv_muladd_vffvf_vf(const size_t n,
			   const float* xvec, const float s, const float *yvec,
			   float* res);
void qn_pp_muladd_vffvf_vf(const size_t n,
			   const float* xvec, const float s, const float *yvec,
			   float* res);
inline void
qn_muladd_vffvf_vf(const size_t n,
		   const float* xvec, const float s, const float *yvec,
		   float* res)
{
    if (qn_math & QN_MATH_PP)
	qn_pp_muladd_vffvf_vf(n, xvec, s, yvec, res);
    else
	qn_nv_muladd_vffvf_vf(n, xvec, s, yvec, res);
}

// Dot product - used by mul_mfvf_vf
float qn_nv_mulsum_vfvf_f(const size_t n, const float* avec,
			  const float* bvec);
float qn_pp_mulsum_vfvf_f(const size_t n, const float* avec,
			  const float* bvec);
inline float
qn_mulsum_vfvf_f(const size_t n, const float* avec, const float* bvec)
{
    if (qn_math & QN_MATH_PP)
	qn_pp_mulsum_vfvf_f(n, avec, bvec);
    else
	qn_nv_mulsum_vfvf_f(n, avec, bvec);
}

//// These are in QN_fltvec_bmul1.cc
// Forward pass
// The internal strided routine
void qn_pp_mulntacc_mfmf_mf_sss(const int M, const int K, const int N,
				const float *const A, const float *const B,
				float *const C,
				const int Astride, const int Bstride,
				const int Cstride);
void qn_nv_mulntacc_mfmf_mf(size_t a_rows, size_t a_cols, size_t b_rows,
			    const float* a, const float* b, float* res);
inline void
qn_pp_mulntacc_mfmf_mf(size_t a_rows, size_t a_cols, size_t b_rows,
		       const float* a, const float* b, float* res)
{
    qn_pp_mulntacc_mfmf_mf_sss(a_rows, a_cols, b_rows,
			       a, b, res, a_cols, a_cols, b_rows);
}

#ifdef QN_HAVE_LIBCBLAS
inline void
qn_bl_mulntacc_mfmf_mf(const int m, const int k, const int n,
		       const float* A, const float* B, float* C)
{
    cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasTrans,
		m, n, k, 1.0f, A, k, B, k, 1.0f, C, n);
}
#endif

inline void
qn_mulntacc_mfmf_mf(size_t a_rows, size_t a_cols, size_t b_rows,
		    const float* a, const float* b, float* res)
{
#ifdef QN_HAVE_LIBCBLAS
    if (qn_math & QN_MATH_BL)
	qn_bl_mulntacc_mfmf_mf(a_rows, a_cols, b_rows, a, b, res);
    else
#endif
    if (qn_math & QN_MATH_PP)
	// Go straight to the internal strided version for speed.
	qn_pp_mulntacc_mfmf_mf(a_rows, a_cols, b_rows,
				   a, b, res);
    else
	qn_nv_mulntacc_mfmf_mf(a_rows, a_cols, b_rows, a, b, res);
}

//// These are in QN_fltvec_bmul2.cc
// Weight update
// Internal strided version
void qn_pp_multnacc_fmfmf_mf_sss(const int M, const int K, const int N, const
 				 float *const A, const float *const B,
				 float *const C,
				 const int Astride, const int Bstride,
				 const int Cstride, const float alpha);
void qn_nv_multnacc_fmfmf_mf(size_t Sk,size_t Sm,size_t Sn, float scale,
			     const float *A,const float *B,float *C);

inline void
qn_pp_multnacc_fmfmf_mf(size_t a_rows, size_t a_cols, size_t b_cols,
			float scale, const float* a, const float* b,
			float* res)
{
    qn_pp_multnacc_fmfmf_mf_sss(a_cols, a_rows, b_cols,
				a, b, res, a_cols, b_cols, b_cols, scale);
}

#ifdef QN_HAVE_LIBCBLAS
inline void
qn_bl_multnacc_fmfmf_mf(const int k, const int m, const int n,
			float alpha, const float* A, const float* B, float* C) 
{
    cblas_sgemm(CblasRowMajor, CblasTrans, CblasNoTrans,
		m, n, k, alpha,	A, m, B, n, 1.0f, C, n);
}
#endif



inline void
qn_multnacc_fmfmf_mf(size_t a_rows, size_t a_cols, size_t b_cols,
		     float scale, const float* a, const float* b,
		     float* res)
{
#ifdef QN_HAVE_LIBCBLAS
    if (qn_math & QN_MATH_BL)
	qn_bl_multnacc_fmfmf_mf(a_rows, a_cols, b_cols, scale, a, b, res);
    else
#endif
    if (qn_math & QN_MATH_PP)
	qn_pp_multnacc_fmfmf_mf(a_rows, a_cols, b_cols, scale, a, b, res);
    else
	qn_nv_multnacc_fmfmf_mf(a_rows, a_cols, b_cols, scale, a, b, res);
}

//// These are in QN_fltvec_bmul3.cc
// Back propagation
// THe optimized internal routine
void qn_pp_mulacc_mfmf_mf_sss(const int M, const int K, const int N, const
			      float 
			 *const A, const float *const B, float *const C, const
			 int Astride, const int Bstride, const int Cstride);

void qn_nv_mulacc_mfmf_mf(size_t Sm,size_t Sk,size_t Sn,
			  const float *A,const float *B,float *C);
inline void
qn_pp_mulacc_mfmf_mf(size_t a_rows, size_t a_cols, size_t b_cols,
		   const float* a, const float* b, float* res)
{
    qn_pp_mulacc_mfmf_mf_sss(a_rows, a_cols, b_cols,
			     a, b, res, a_cols, b_cols, b_cols);
}
#ifdef QN_HAVE_LIBCBLAS
inline void
qn_bl_mulacc_mfmf_mf(const int 	m, const int k, const int n, 
		     const float* A, const float* B, float* C) 
{
    cblas_sgemm(CblasRowMajor,CblasNoTrans, CblasNoTrans,
		m, n, k, 1.0f, A, k, B, n, 1.0f, C, n);
}
#endif

inline void
qn_mulacc_mfmf_mf(size_t a_rows, size_t a_cols, size_t b_cols,
		  const float* a, const float* b, float* res)
{
#ifdef QN_HAVE_LIBCBLAS
    if (qn_math & QN_MATH_BL)
	qn_bl_mulacc_mfmf_mf(a_rows, a_cols, b_cols, a, b, res)	;
    else
#endif
    if (qn_math & QN_MATH_PP)
	qn_pp_mulacc_mfmf_mf(a_rows, a_cols, b_cols, a, b, res)	;
    else
	qn_nv_mulacc_mfmf_mf(a_rows, a_cols, b_cols, a, b, res)	;
}

void qn_nv_mul_mfmf_mf(size_t Sm,size_t Sk,size_t Sn,
		       const float *A,const float *B,float *C);
inline void
qn_pp_mul_mfmf_mf(size_t a_rows, size_t a_cols, size_t b_cols,
		  const float* a, const float* b, float* res)
{
    // A hack because we do not have a non-accumulating version
    qn_copy_f_vf(a_rows * b_cols, 0.0f, res);
    qn_pp_mulacc_mfmf_mf_sss(a_rows, a_cols, b_cols,
			     a, b, res, a_cols, b_cols, b_cols);
}

#ifdef QN_HAVE_LIBCBLAS
inline void
qn_bl_mul_mfmf_mf(const int m, const int k, const int n, 
		  const float* A, const float* B, float* C) 
{
    cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
		m, n, k, 1.0, A, k, B, n, 0.0f, C, n);
}
#endif

inline void
qn_mul_mfmf_mf(size_t a_rows, size_t a_cols, size_t b_cols,
	       const float* a, const float* b, float* res)
{
#ifdef QN_HAVE_LIBCBLAS
    if (qn_math & QN_MATH_BL)
	qn_bl_mul_mfmf_mf(a_rows, a_cols, b_cols, a, b, res);
    else
#endif
    if (qn_math & QN_MATH_PP)
	qn_pp_mul_mfmf_mf(a_rows, a_cols, b_cols, a, b, res);
    else
	qn_nv_mul_mfmf_mf(a_rows, a_cols, b_cols, a, b, res);
}


//// *** Inline routines ***

// Return true if a float is not a number
inline QNBool
qn_isnan_f_b(float f)
{
    /* With IEEE FP, a NaN does not equal anything */
    return ((f==f) ? QN_FALSE : QN_TRUE);
}

static union
{
	double d;
	struct 
	{
		int j,i;
	} n;
} qn_d2i;

// Here we have macros for really hacky but quick exponential function.
// This was taken from Chris Oei's version and tidied up a little.
// I don't know anything about the origin, but it appears to be relying
// on the fact that storing an int in the exponent of a double gives
// you an exponential in the higher bits with the lower bits being
// used to add a straight line fit.
// Note that, due to the use of the QN_WORDS_BIGENDIAN macro, only works on
// machines where the words-in-a-double ordering  is the same as the
// bytes-in-a-word ordering.  There are exceptions, e.g. the Vax!
// It also relies on the static qn_d2i being initialized to zero.

#define QN_EXP_A (1048576/M_LN2)
#define QN_EXP_C 60801
#ifdef QN_WORDS_BIGENDIAN
#define QN_EXPQ(y) (qn_d2i.n.j = (int) (QN_EXP_A*(y)) + (1072693248 - QN_EXP_C), qn_d2i.d)
#else
#define QN_EXPQ(y) (qn_d2i.n.i = (int) (QN_EXP_A*(y)) + (1072693248 - QN_EXP_C), qn_d2i.d)
#endif

inline float
qn_pp_sigmoid_f_f(float x)
{
    return 1.0f/(1.0f + QN_EXPQ(-x));
}

inline float
qn_nv_sigmoid_f_f(float x)
{
#ifdef QN_HAVE_EXPF
    return 1.0f/(1.0f + expf(-x));
#else
    return 1.0f/(1.0f + exp(-x));
#endif
}

inline float
qn_sigmoid_f_f(float x)
{
    float r;

    if (qn_math & QN_MATH_PP)
	r = qn_pp_sigmoid_f_f(x);
    else
	r = qn_nv_sigmoid_f_f(x);
    return r;
}

inline void
qn_sigmoid_vf_vf(size_t n, const float* in_vec, float* out_vec)
{
    if (qn_math & QN_MATH_PP)
	qn_pp_sigmoid_vf_vf(n, in_vec, out_vec);
    else
	qn_nv_sigmoid_vf_vf(n, in_vec, out_vec);
}

inline void
qn_softmax_vf_vf(size_t n, const float* in_vec, float* out_vec)
{
    if (qn_math & QN_MATH_PP)
	qn_pp_softmax_vf_vf(n, in_vec, out_vec);
    else
	qn_nv_softmax_vf_vf(n, in_vec, out_vec);
}

inline float
qn_ierf_f_f(float x)
{

     float last = 1.0f;		/* Current approximate ierf val */
     float curr = 0.0f;		/* Previous approximate ierf val */
     float erfap;		/* Erf() of current approximate ierf */
     float slope;		/* Slope at current point on ierf curve */
     const float DELTA = (1.0 / (1 << 10)); /* Delta use to calc. slope */
     const float LIMIT = (1.0 / (1 << 20)); /* Stop when error this small */
     
     while(fabs(last - curr) > LIMIT) {
	  erfap = erf(curr);
	  slope = (erf(curr + DELTA) - erf(curr)) / DELTA;
	  
	  last = curr;
	  curr += (x - erfap) / slope;
     }
     return(curr);
}

inline float
qn_erf_f_f(float x)
{
    return erf(x);
}

//************************************************/
//** Cambridge "pre-file" conversion functions ***/
//************************************************/
//
// The "pre" format is used by the Cambridge speech group to store
// speech features in a file efficiently.  The format uses bytes to encode
// a floating point feature value.
//

inline QNInt8
qn_htopre8_f_i8(float val)
{
    QNInt32 res;

    /* QN_SQRT1_2 = sqrt(1/2) */
    res = (QNInt32) floor( (erf(val * QN_SQRT1_2)+1.0) * 128.0 );
    if (res>QN_UINT8_MAX)
	res = QN_UINT8_MAX;
    assert(res>=0);
    return ((QNInt8) res);
}

inline float
qn_pre8toh_i8_f(QNInt8 bits)
{
    float ret;			/* Return value */
    float* table;		/* Pointer to the start of the lookup table */
    
    table = qn_pre8table_vf();	/* Point to the table for converting pre8
				   bytes to floating point */
    ret = table[(QNUInt8) bits]; /* Perform the lookup */

    return ret;
}

//************************************************/
//** Cambridge "LNA file" conversion functions ***/
//************************************************/
//
// The "LNA" format is used by the Cambridge and ICSI speech groups to store
// probabilities efficiently.
// To convert a probability to an 8 bit LNA value, the following formula
// is used:
//
//       lna8 = trunc(-24 * log(x + VERY_SMALL))
//

#define QN_LNA8_VERY_SMALL (1e-37)
#define QN_LNA8_SCALE (24.0)

inline QNInt8
qn_htolna8_f_i8(float val)
{
    QNInt32 res;

    res = (int) floor( - QN_LNA8_SCALE * log( val + QN_LNA8_VERY_SMALL ) ) ;
    if (res>QN_UINT8_MAX)
	res = QN_UINT8_MAX;
    if (res<0)
	res = 0;
    return ((QNInt8) res);
}

inline float
qn_lna8toh_i8_f(QNInt8 lna)
{
    float ret;			/* Return value */
    float* table;		/* Pointer to the start of the lookup table */
    
    table = qn_lna8table_vf();	/* Point to the table for converting lna8
				   bytes to floating point */
    ret = table[(QNUInt8) lna]; /* Perform the lookup */

    return ret;
}

// btoh is for "bigendian to host" format conversion.  Note that on big
// endian systems this is a no-op.

inline void
qn_btoh_vf_vf(size_t len, const float* from, float* to)
{
    qn_btoh_vi32_vi32(len, (const QNInt32*) from, (QNInt32*) to);
}

inline void
qn_htob_vf_vf(size_t len, const float* from, float* to)
{
    qn_htob_vi32_vi32(len, (const QNInt32*) from, (QNInt32*) to);
}


inline void
qn_copy_d_vd(size_t len, double from, double* to)
{
    size_t i;

    for (i=0; i<len; i++)
	*to++ = from;
}

inline void
qn_copy_vf_vf(size_t len, const float* from, float* to)
{
    size_t i;

    for (i=0; i<len; i++)
	*to++ = *from++;
}

inline void
qn_copy_vd_vd(size_t len, const double* from, double* to)
{
    size_t i;

    for (i=0; i<len; i++)
	*to++ = *from++;
}

inline void
qn_conv_vf_vd(size_t len, const float* a, double* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (double) *a++;
}

// Convert a vector of doubles to a vector of floats using rounding.

inline void
qn_conv_vd_vf(size_t len, const double* a, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (float) *a++;
}

inline void
qn_acc_vf_vf(size_t n, const float *in_vec, float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] += in_vec[i];
}

inline void
qn_acc_vd_vd(size_t n, const double *in_vec, double *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] += in_vec[i];
}

inline void
qn_neg_vf_vf(size_t len, const float* from, float* to)
{
    size_t i;

    for (i=0; i<len; i++)
    {
	*to++ = - (*from++);
    }
}

inline void
qn_sqrt_vf_vf(size_t len, const float* a, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
    {
#ifdef QN_HAVE_SQRTF
	*res++ = sqrtf(*a++);
#else
	*res++ = sqrt(*a++);
#endif
    }
}

inline void
qn_sqr_vf_vf(size_t len, const float* a, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
    {
	*res++ = (*a) * (*a);
	a++;
    }
}

inline void
qn_sqr_vd_vd(size_t len, const double* a, double* res)
{
    size_t i;

    for (i=0; i<len; i++)
    {
	*res++ = (*a) * (*a);
	a++;
    }
}

inline void
qn_recip_vf_vf(size_t len, const float* a, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
    {
	*res++ = 1.0f/(*a++);
    }
}

inline void
qn_add_vfvf_vf(size_t n,
	       const float *in_vec1, const float *in_vec2,
	       float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] + in_vec2[i];
}

inline void
qn_sub_vff_vf
(size_t len, const float* a, float b, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (*a++) - b;
}

inline void
qn_sub_vfvf_vf(size_t n,
	       const float *in_vec1, const float *in_vec2,
	       float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] - in_vec2[i];
}

inline void
qn_sub_vdvd_vd(size_t n,
	       const double *in_vec1, const double *in_vec2,
	       double *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] - in_vec2[i];
}

inline void
qn_mul_vff_vf(size_t len, const float* a, float b, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (*a++) * b;
}

inline void
qn_mul_vdd_vd(size_t len, const double* a, double b, double* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = (*a++) * b;
}

inline void
qn_div_fvf_vf(size_t len, float a, const float* b, float* res)
{
    size_t i;

    for (i=0; i<len; i++)
	*res++ = a / (*b++);
}

inline void
qn_mul_vfvf_vf(size_t n,
	       const float *in_vec1, const float *in_vec2,
	       float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] * in_vec2[i];
}

inline void
qn_div_vfvf_vf(size_t n,
	       const float *in_vec1, const float *in_vec2,
	       float *res_vec)
{
    size_t i;
    for (i=0; i<n; i++)
        res_vec[i] = in_vec1[i] / in_vec2[i];
}

inline void
qn_mulacc_vff_vf(size_t n, const float *in, float scale, float *acc)
{
    size_t i;
    for (i=0; i<n; i++)
        acc[i] += scale * in[i];
}

inline void
qn_dsigmoid_vf_vf(size_t n, const float* in_vec, float* out_vec)
{
    size_t i;

    for (i=0; i<n; i++)
    {
        const float y = in_vec[i];

	out_vec[i] = (1.0f - y) * y;
    }
}


inline void
qn_maxmin_vf_ff(size_t n, const float *vec, float *maxp, float *minp)
{
    if (n>0)
    {
        float max = vec[0];
        float min = vec[0];
        size_t i;
        for (i=1; i<n; i++)
        {
            const float elem = vec[i];
            if (elem > max)
                max = elem;
            else if (elem < min)
                min = elem;
        }

        *maxp = max;
        *minp = min;
    }
}

#endif /* #ifndef QN_fltvec_h_INCLUDED */

